#!/usr/bin/env python

"""
HTML serialiser.

Copyright (C) 2017, 2018, 2019, 2021, 2022,
              2023 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from moinformat.serialisers.common import escape_attr, escape_text, Serialiser
from moinformat.tree.moin import LinkLabel, LinkParameter, Text
from moinformat.utils.links import parse_link_target

class HTMLSerialiser(Serialiser):

    "Serialisation of the page."

    input_formats = ["moin", "wiki"]
    formats = ["html"]

    # Support methods.

    list_tags = {
        "i" : "lower-roman",
        "I" : "upper-roman",
        "a" : "lower-latin",
        "A" : "upper-latin",
        }

    def _get_list_tag(self, marker):
        if marker:
            if marker[0].isdigit():
                return "ol", "decimal"
            style_type = self.list_tags.get(marker[0])
            if style_type:
                return "ol", style_type

        return "ul", None

    def _link(self, target, nodes, tag, attr):
        link = self.linker and self.linker.translate(target) or None

        self.out('<%s %s="%s"' % (tag, attr, escape_attr(link.get_target())))

        # Provide link parameters as attributes.

        if nodes:
            for node in nodes:
                if isinstance(node, LinkParameter):
                    self.out(" ")
                    node.visit(self)

        # Close the tag if an image.

        if tag == "img":
            self.out(" />")

        # Provide the link label if specified. Otherwise, use a generated
        # default for the label.

        else:
            self.out(">")

            for node in nodes or []:
                if isinstance(node, LinkLabel):
                    node.visit(self)
                break
            else:
                self.out(escape_text(link.get_label()))

            self.out("</%s>" % tag)

    def _region_tag(self, type):

        # NOTE: Need to support types in general.

        type = type and type.split()[0]

        if type == "inline":
            return "tt"
        elif type in (None, "python"):
            return "pre"
        else:
            return "span"

    # Node handler methods.

    def region(self, region):
        tag = self._region_tag(region.type)

        # Generate attributes, joining them when preparing the tag.

        attrs = []
        attr = attrs.append

        if region.level:
            attr("level-%d" % region.level)

        if region.indent:
            attr("indent-%d" % region.indent)

        # NOTE: Encode type details for CSS.

        attr("type-%s" % escape_attr(region.type or "opaque"))

        # Inline regions must preserve "indent" as space in the text.

        if region.type == "inline" and region.indent:
            self.out(" " * region.indent)

        self.out("<%s class='%s'>" % (tag, " ".join(attrs)))

        # Serialise the region content.

        self.visit_region(region)

        # End the region with the previous serialiser.

        self.out("</%s>" % tag)

    # Block node methods.

    def block(self, block):
        self.out("<p>")
        self.container(block)
        self.out("</p>")

    def defitem(self, defitem):
        self.out("<dd>")
        self.container(defitem)
        self.out("</dd>")

    def defterm(self, defterm):
        self.out("<dt>")
        self.container(defterm)
        self.out("</dt>")

    def fontstyle(self, fontstyle):
        if fontstyle.emphasis:
            self.out("<em>")
        elif fontstyle.strong:
            self.out("<strong>")
        self.container(fontstyle)
        if fontstyle.emphasis:
            self.out("</em>")
        elif fontstyle.strong:
            self.out("</strong>")

    def heading(self, heading):
        self.out("<h%d id='%s'>" % (
            heading.level,
            escape_attr(self.linker.make_id(heading.identifier))))
        self.container(heading)
        self.out("</h%d>" % heading.level)

    def larger(self, larger):
        self.out("<big>")
        self.container(larger)
        self.out("</big>")

    def list(self, list):
        tag, style_type = self._get_list_tag(list.marker)
        style = style_type and \
            ' style="list-style-type: %s"' % escape_attr(style_type) or ""
        start = style_type and \
            list.num is not None and ' start="%s"' % escape_attr(list.num) or ""
        self.out("<%s%s%s>" % (tag, style, start))
        self.container(list)
        self.out("</%s>" % tag)

    def listitem(self, listitem):
        self.out("<li>")
        self.container(listitem)
        self.out("</li>")

    def macro(self, macro):

        # Special case of a deliberately unexpanded macro.

        if macro.nodes is None:
            return

        tag = macro.inline and "span" or "div"
        self.out("<%s class='macro %s'>" % (tag, escape_text(macro.name)))

        # Fallback case for when macros are not replaced.

        if not macro.nodes:
            self.out(escape_text("<<"))
            self.out("<span class='name'>%s</span>" % escape_text(macro.name))
            if macro.args:
                self.out("(")
            first = True
            for arg in macro.args:
                if not first:
                    self.out(",")
                self.out("<span class='arg'>%s</span>" % escape_text(arg))
                first = False
            if macro.args:
                self.out(")")
            self.out(escape_text(">>"))

        # Produce the expanded macro content.

        else:
            self.container(macro)

        tag = macro.inline and "span" or "div"
        self.out("</%s>" % tag)

    def monospace(self, monospace):
        self.out("<tt>")
        self.container(monospace)
        self.out("</tt>")

    def smaller(self, smaller):
        self.out("<small>")
        self.container(smaller)
        self.out("</small>")

    def strikethrough(self, strikethrough):
        self.out("<del>")
        self.container(strikethrough)
        self.out("</del>")

    def subscript(self, subscript):
        self.out("<sub>")
        self.container(subscript)
        self.out("</sub>")

    def superscript(self, superscript):
        self.out("<sup>")
        self.container(superscript)
        self.out("</sup>")

    def table(self, table):
        self.out("<table>")
        self.container(table)
        self.out("</table>")

    def table_cell(self, table_cell):
        self.out("<td")

        # Handle the attributes separately from their container.

        if table_cell.attrs and not table_cell.attrs.empty():
            for attr in table_cell.attrs.nodes:
                attr.visit(self)

        self.out(">")
        self.container(table_cell)
        self.out("</td>")

    def table_row(self, table_row):
        self.out("<tr>")
        self.container(table_row)
        self.out("</tr>")

    def underline(self, underline):
        self.out("<span style='text-decoration: underline'>")
        self.container(underline)
        self.out("</span>")

    # Inline node methods.

    def anchor(self, anchor):
        self.out("<a name='%s' />" % escape_attr(self.linker.make_id(anchor.target)))

    def break_(self, break_):
        pass

    def comment(self, comment):
        pass

    def directive(self, directive):

        # Obtain a blank value if the value is missing.

        name, text = (directive.directive.split(None, 1) + [""])[:2]

        # Produce a readable redirect.

        if name.lower() == "redirect":
            self.start_block()

            # Process the redirect argument as a link target, producing a link
            # element.

            target = parse_link_target(text, self.metadata)
            self._link(target, [LinkLabel([Text(text)])], "a", "href")

            self.end_block()

    def linebreak(self, linebreak):
        self.out("<br />")

    def link(self, link):
        self._link(link.target, link.nodes, "a", "href")

    def link_label(self, link_label):
        self.container(link_label)

    def link_parameter(self, link_parameter):
        s = link_parameter.text_content()
        key_value = s.split("=", 1)

        if len(key_value) == 1:
            self.out(key_value[0])
        else:
            key, value = key_value
            self.out("%s='%s'" % (key, escape_attr(value)))

    def nbsp(self, nbsp):
        self.out("&nbsp;")

    def rule(self, rule):
        self.out("<hr style='height: %dpt' />" % min(rule.height, 10))

    def table_attrs(self, table_attrs):

        # Skip the attributes in their original form.

        pass

    def table_attr(self, table_attr):
        self.out(" %s%s" % (
            escape_text(table_attr.name),
            table_attr.value is not None and
            "='%s'" % escape_attr(table_attr.value) or ""))

    def text(self, text):
        self.out(escape_text(text.s))

    def transclusion(self, transclusion):
        self._link(transclusion.target, transclusion.nodes, "img", "src")

    def verbatim(self, verbatim):
        self.out(escape_text(verbatim.text))

serialiser = HTMLSerialiser

# vim: tabstop=4 expandtab shiftwidth=4
