#!/usr/bin/env python

"""
Moin wiki format document tree nodes.

Copyright (C) 2017, 2018, 2019, 2020, 2021, 2022,
              2023 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

class Container:

    "A container of document nodes."

    def __init__(self, nodes):
        self.nodes = nodes

        # In principle, allow blocks within containers. Some nodes may forbid
        # them to simplify the document structure.

        self.allow_blocks = True

    def append(self, node):
        self.nodes.append(node)

    def append_many(self, nodes):
        for node in nodes:
            self.append(node)

    def add(self, node):
        last = self.node(-1)
        if last and last.empty():
            self.nodes[-1] = node
        else:
            self.append(node)

    def append_inline(self, node):

        "Append 'node' inline within the appropriate container."

        n = self.append_point()

        # Redirect the addition if another container is to accept the node.

        if n is not self:
            n.append_inline(node)

        # Otherwise, append within this container.

        else:
            n.append(node)

    def append_inline_many(self, nodes):
        for node in nodes:
            self.append_inline(node)

    def append_point(self):

        "Return the container to which inline nodes are added."

        last = self.node(-1)

        if isinstance(last, Block):
            return last
        else:
            return self

    def empty(self):
        return not self.nodes

    def insert_after(self, old, new):

        "Insert after 'old' in the children the 'new' node."

        index = self.nodes.index(old)
        self.nodes.insert(index + 1, new)

    def node(self, index):
        try:
            return self.nodes[index]
        except IndexError:
            return None

    def normalise(self):

        "Combine adjacent text nodes."

        nodes = self.nodes
        self.nodes = []
        text = None

        for node in nodes:

            # Open a text node or merge text into an open node.

            if isinstance(node, Text):
                if not text:
                    text = node
                else:
                    text.merge(node)

            # Close any open text node and append the current node.

            else:
                if text:
                    self.append(text)
                    text = None
                self.append(node)

        # Add any open text node.

        if text:
            self.append(text)

    def remove(self, node):

        "Remove 'node' from the children."

        self.nodes.remove(node)

    def replace(self, old, new):

        "Replace 'old' with 'new' in the children."

        i = self.nodes.index(old)
        self.nodes[i] = new

    def split_at(self, node):

        """
        Split the container at 'node', returning a new container holding the
        nodes following 'node' that are moved from this container.
        """

        i = self.nodes.index(node)
        following = self.__class__(self.nodes[i+1:])

        # Remove the node and the following parts from this container.

        del self.nodes[i:]
        return following

    def text_content(self):

        """
        Return a string containing the content of text nodes within this
        container.
        """

        l = []

        for node in self.nodes:
            if isinstance(node, Text):
                l.append(node.s)
            elif isinstance(node, Container):
                l.append(node.text_content())

        return "".join(l)

    def whitespace_only(self):

        "Return whether the container provides only whitespace text."

        return not self.text_content().strip()

class Region(Container):

    "A region of the page."

    def __init__(self, nodes, level=0, indent=0, type=None, args=None,
                 transparent=True, extra=None):
        Container.__init__(self, nodes)
        self.level = level
        self.indent = indent
        self.type = type
        self.args = args
        self.transparent = transparent
        self.extra = extra

    def append_point(self):

        "Return the container to which inline nodes are added."

        if self.transparent:
            return self.nodes[-1]
        else:
            return self

    def have_end(self, s):
        return self.level and s.startswith("}") and self.level == len(s)

    def __repr__(self):
        return "Region(%r, %r, %r, %r, %r, %r, %r)" % (self.nodes, self.level,
            self.indent, self.type, self.args, self.transparent, self.extra)

    def visit(self, visitor):
        visitor.region(self)



# Block nodes.

class Block(Container):

    "A block in the page."

    def __repr__(self):
        return "Block(%r)" % self.nodes

    def visit(self, visitor):
        visitor.block(self)

class DefItem(Container):

    "A definition item."

    def __init__(self, nodes, pad, extra):
        Container.__init__(self, nodes)
        self.pad = pad
        self.extra = extra

    def __repr__(self):
        return "DefItem(%r, %r, %r)" % (self.nodes, self.pad, self.extra)

    def visit(self, visitor):
        visitor.defitem(self)

class DefTerm(Container):

    "A definition term."

    def __init__(self, nodes, pad, extra=""):
        Container.__init__(self, nodes)
        self.pad = pad
        self.extra = extra

    def __repr__(self):
        return "DefTerm(%r, %r, %r)" % (self.nodes, self.pad, self.extra)

    def visit(self, visitor):
        visitor.defterm(self)

class FontStyle(Container):

    "Emphasised and/or strong text."

    def __init__(self, nodes, emphasis=False, strong=False):
        Container.__init__(self, nodes)
        self.emphasis = emphasis
        self.strong = strong

    def close_emphasis(self):
        if self.strong:
            span = FontStyle(self.nodes, emphasis=True)
            self.nodes = [span]
            self.emphasis = False
        return self.strong

    def close_strong(self):
        if self.emphasis:
            span = FontStyle(self.nodes, strong=True)
            self.nodes = [span]
            self.strong = False
        return self.emphasis

    def __repr__(self):
        return "FontStyle(%r, %r, %r)" % (self.nodes, self.emphasis, self.strong)

    def visit(self, visitor):
        visitor.fontstyle(self)

class Heading(Container):

    "A heading."

    def __init__(self, nodes, level, start_extra="", start_pad="", end_pad="", end_extra="",
                 identifier=None):
        Container.__init__(self, nodes)
        self.level = level
        self.start_extra = start_extra
        self.start_pad = start_pad
        self.end_pad = end_pad
        self.end_extra = end_extra
        self.identifier = identifier

    def __repr__(self):
        return "Heading(%r, %d, %r, %r, %r, %r, %r)" % (
            self.nodes, self.level, self.start_extra, self.start_pad,
            self.end_pad, self.end_extra, self.identifier)

    def visit(self, visitor):
        visitor.heading(self)

class LinkLabel(Container):

    "A link or transclusion label."

    def __repr__(self):
        return "LinkLabel(%r)" % self.nodes

    def visit(self, visitor):
        visitor.link_label(self)

class LinkParameter(Container):

    "A link or transclusion parameter."

    def __repr__(self):
        return "LinkParameter(%r)" % self.nodes

    def visit(self, visitor):
        visitor.link_parameter(self)

class List(Container):

    "A list."

    def __init__(self, nodes):
        Container.__init__(self, nodes)
        self.init()

    def init(self):
        self.first = first = self.nodes and self.nodes[0] or None
        self.indent = first and first.indent
        self.marker = first and first.marker
        self.num = first and first.num

    def __repr__(self):
        return "List(%r)" % self.nodes

    def visit(self, visitor):
        if not self.first:
            self.init()
        visitor.list(self)

class ListItem(Container):

    "A list item."

    def __init__(self, nodes, indent, marker, space, num):
        Container.__init__(self, nodes)
        self.indent = indent
        self.marker = marker
        self.space = space
        self.num = num

        # Forbid blocks within list items for simpler structure.

        self.allow_blocks = False

    def __repr__(self):
        return "ListItem(%r, %r, %r, %r, %r)" % (self.nodes, self.indent, self.marker, self.space, self.num)

    def visit(self, visitor):
        visitor.listitem(self)

class Table(Container):

    "A table."

    def __repr__(self):
        return "Table(%r)" % self.nodes

    def visit(self, visitor):
        visitor.table(self)

class TableAttrs(Container):

    "A collection of table attributes."

    def __init__(self, nodes):
        Container.__init__(self, nodes)

        # Parsing state flags, inconsequential to any final document tree.
        # If incomplete remains set, the attributes are discarded.

        self.incomplete = True
        self.found_cell = False

    def __repr__(self):
        return "TableAttrs(%r)" % self.nodes

    def visit(self, visitor):
        visitor.table_attrs(self)

class TableCell(Container):

    "A table cell."

    def __init__(self, nodes, attrs=None, leading="", padding=""):
        Container.__init__(self, nodes)
        self.attrs = attrs
        self.leading = leading
        self.padding = padding

    def __repr__(self):
        return "TableCell(%r, %r, %r, %r)" % (self.nodes, self.attrs,
                                              self.leading, self.padding)

    def visit(self, visitor):
        visitor.table_cell(self)

class TableRow(Container):

    "A table row."

    def __init__(self, nodes, trailing="", leading="", padding=""):
        Container.__init__(self, nodes)
        self.trailing = trailing
        self.leading = leading
        self.padding = padding

    def __repr__(self):
        return "TableRow(%r, %r, %r, %r)" % (self.nodes, self.trailing,
                                             self.leading, self.padding)

    def visit(self, visitor):
        visitor.table_row(self)



# Inline nodes with children.

class Inline(Container):

    "Generic inline formatting."

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.nodes)

class Larger(Inline):

    "Larger text."

    def visit(self, visitor):
        visitor.larger(self)

class Link(Container):

    "Link details."

    def __init__(self, nodes, target):
        Container.__init__(self, nodes)
        self.target = target

    def __repr__(self):
        return "Link(%r, %r)" % (self.nodes, self.target)

    def visit(self, visitor):
        visitor.link(self)

class Macro(Container):

    "Macro details."

    def __init__(self, name, args, parent, region, nodes=None, inline=True):
        Container.__init__(self, nodes or [])
        self.name = name
        self.args = args
        self.parent = parent
        self.region = region
        self.inline = inline

    def __repr__(self):
        return "Macro(%r, %r, %r, %r, %r, %r)" % (self.name, self.args,
                                                  self.parent, self.region,
                                                  self.nodes, self.inline)

    def visit(self, visitor):
        visitor.macro(self)

class Monospace(Inline):

    "Monospaced text."

    def visit(self, visitor):
        visitor.monospace(self)

class Smaller(Inline):

    "Smaller text."

    def visit(self, visitor):
        visitor.smaller(self)

class Strikethrough(Inline):

    "Crossed-visitor text."

    def visit(self, visitor):
        visitor.strikethrough(self)

class Subscript(Inline):

    "Subscripted text."

    def visit(self, visitor):
        visitor.subscript(self)

class Superscript(Inline):

    "Superscripted text."

    def visit(self, visitor):
        visitor.superscript(self)

class Transclusion(Container):

    "Transclusion details."

    def __init__(self, nodes, target):
        Container.__init__(self, nodes)
        self.target = target

    def __repr__(self):
        return "Transclusion(%r, %r)" % (self.nodes, self.target)

    def visit(self, visitor):
        visitor.transclusion(self)

class Underline(Inline):

    "Underlined text."

    def visit(self, visitor):
        visitor.underline(self)



# Nodes without children.

class Node:

    "A document node without children."

    def empty(self):
        return False

class Anchor(Node):

    "Anchor details."

    def __init__(self, target):
        self.target = target

    def __repr__(self):
        return "Anchor(%r)" % self.target

    def visit(self, visitor):
        visitor.anchor(self)

class Break(Node):

    "A paragraph break."

    def __repr__(self):
        return "Break()"

    def visit(self, visitor):
        visitor.break_(self)

class Comment(Node):

    "A comment."

    def __init__(self, comment, extra):
        self.comment = comment
        self.extra = extra

    def __repr__(self):
        return "Comment(%r, %r)" % (self.comment, self.extra)

    def visit(self, visitor):
        visitor.comment(self)

class Directive(Node):

    "A processing directive."

    def __init__(self, directive, extra):
        self.directive = directive
        self.extra = extra

    def __repr__(self):
        return "Directive(%r, %r)" % (self.directive, self.extra)

    def visit(self, visitor):
        visitor.directive(self)

class LineBreak(Node):

    "A line break within a block."

    def __repr__(self):
        return "LineBreak()"

    def visit(self, visitor):
        visitor.linebreak(self)

class NonBreakingSpace(Node):

    "A non-breaking space within a block."

    def __repr__(self):
        return "NonBreakingSpace()"

    def visit(self, visitor):
        visitor.nbsp(self)

class Rule(Node):

    "A horizontal rule."

    def __init__(self, height):
        self.height = height

    def __repr__(self):
        return "Rule(%d)" % self.height

    def visit(self, visitor):
        visitor.rule(self)

class TableAttr(Node):

    "A table attribute."

    def __init__(self, name, value=None, concise=False, quote=None):
        self.name = name
        self.value = value
        self.concise = concise
        self.quote = quote

    def __repr__(self):
        return "TableAttr(%r, %r, %r, %r)" % (self.name, self.value, self.concise, self.quote)

    def visit(self, visitor):
        visitor.table_attr(self)

class Text(Node):

    "A text node."

    def __init__(self, s):
        self.s = s

    def empty(self):
        return not self.s

    def multiline(self):
        return "\n" in self.s

    def merge(self, text):
        self.s += text.s

    def __repr__(self):
        return "Text(%r)" % self.s

    def visit(self, visitor):
        visitor.text(self)

class Verbatim(Node):

    "Verbatim text."

    def __init__(self, text):
        self.text = text

    def __repr__(self):
        return "Verbatim(%r)" % self.text

    def visit(self, visitor):
        visitor.verbatim(self)

# vim: tabstop=4 expandtab shiftwidth=4
